"""
Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
from proton.keyring_linux.core import KeyringBackendLinux

logger = logging.getLogger(__name__)


class LibsecretKeyringBackend(KeyringBackendLinux):
    """
    Libsecret backend using high-level password APIs (no explicit Service).
    https://gnome.pages.gitlab.gnome.org/libsecret/libsecret-python-examples.html
    """

    _secret_cache = None

    # We are using the same schema and application name
    # as the other keyring backend to be compatible
    SCHEMA = "org.freedesktop.Secret.Generic"
    APPLICATION_NAME = "Python keyring library"

    @classmethod
    def _get_priority(cls):
        return 4.0  # lower than secretservice

    @staticmethod
    def _import_secret():
        import gi  # pylint: disable=import-outside-toplevel,import-error
        gi.require_version('Secret', '1')
        from gi.repository import Secret  # pylint: disable=import-outside-toplevel,import-error
        return Secret

    @classmethod
    def _get_secret(cls):
        # The import is not top level because the app shouldn't crash
        # when the optional dependency is not available.
        # It's only imported the first time it's used and then cached
        if cls._secret_cache is None:
            cls._secret_cache = cls._import_secret()
        return cls._secret_cache

    @classmethod
    def _validate(cls):
        try:
            secret = cls._get_secret()
        except ValueError:
            logger.warning("Libsecret not available")
            return False
        test_schema = secret.Schema.new(
            "test",
            secret.SchemaFlags.NONE,
            {"key": secret.SchemaAttributeType.STRING}
        )
        secret.password_lookup_sync(test_schema, {"key": "test"}, None)
        logger.info("Using libsecret backend")
        return True

    def __init__(self):

        secret = self._get_secret()
        # These keys are used to find the keyring entry later.
        # They are being kept compatible with existing keys from secret service backend
        # so that the backends can be used interchangeably without losing user data.
        self.schema = secret.Schema.new(
            LibsecretKeyringBackend.SCHEMA,
            secret.SchemaFlags.NONE,  # pylint: disable=no-member
            {
                "application": secret.SchemaAttributeType.STRING,
                "service": secret.SchemaAttributeType.STRING,
                "username": secret.SchemaAttributeType.STRING
            }
        )

        super().__init__(self)

    def get_password(self, service_name, key):
        "Retrieve a password"
        secret = self._get_secret()
        result = secret.password_lookup_sync(
            self.schema,
            {
                "application": LibsecretKeyringBackend.APPLICATION_NAME,
                "service": service_name,
                "username": key
            },
            None
        )
        if result is None:
            return None
        return result

    def set_password(self, service_name, key, password):
        """Store a password"""
        secret = self._get_secret()
        secret.password_store_sync(
            self.schema,
            {
                "application": LibsecretKeyringBackend.APPLICATION_NAME,
                "service": service_name,
                "username": key
            },
            None,
            key,  # this is the title of the keyring entry
            password,
            None
        )

    def delete_password(self, service_name, key):
        """Delete a password"""
        secret = self._get_secret()
        secret.password_clear_sync(
            self.schema,
            {
                "application": LibsecretKeyringBackend.APPLICATION_NAME,
                "service": service_name,
                "username": key
            },
            None
        )



"""
Module for all linux keyring backends.


The public interface and the functionality that's common to all supported
VPN connection backends is defined in this module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import json
import logging

import keyring
from proton.keyring._base import Keyring
from proton.keyring.exceptions import KeyringLocked, KeyringError

logger = logging.getLogger(__name__)


class KeyringBackendLinux(Keyring):  # pylint: disable=too-few-public-methods
    """Keyring linux backend.

    All backend implementations should derive from this class, so methods
    can be reused, unless there are backend specific approaches.
    """
    KEYRING_SERVICE = "Proton"

    def __init__(self, keyring_backend):
        super().__init__()
        self.__keyring_backend = keyring_backend

    # pylint: disable=duplicate-code
    def _get_item(self, key):
        try:
            stored_data = self.__keyring_backend.get_password(  # pylint: disable=duplicate-code
                self.KEYRING_SERVICE,
                key
            )
        except keyring.errors.KeyringLocked as excp:
            logging.info("Keyring locked while getting")
            raise KeyringLocked("Keyring is locked") from excp
        except keyring.errors.KeyringError as excp:
            logging.exception("Keyring error while getting")
            raise KeyringError(excp) from excp

        # Since we're borrowing the dict interface,
        # be consistent and throw a KeyError if it doesn't exist
        if stored_data is None:
            raise KeyError(key)

        try:
            return json.loads(stored_data)
        except json.JSONDecodeError as excp:
            # Delete data (it's invalid anyway)
            self._del_item(key)
            raise KeyError(key) from excp

    def _del_item(self, key):
        try:
            self.__keyring_backend.delete_password(self.KEYRING_SERVICE, key)
        except keyring.errors.PasswordDeleteError as excp:
            logging.exception("Unable to delete entry from keyring")
            raise KeyError(key) from excp
        except keyring.errors.KeyringError as excp:
            logging.exception("Keyring error while deleting")
            raise KeyringError(excp) from excp

    def _set_item(self, key, value):
        json_data = json.dumps(value)
        try:
            self.__keyring_backend.set_password(
                self.KEYRING_SERVICE,
                key,
                json_data
            )
        except keyring.errors.PasswordSetError as excp:
            logging.info("Unable to set value to keyring")
            raise KeyError(excp) from excp
        except keyring.errors.KeyringError as excp:
            logging.exception("Keyring error while setting")
            raise KeyringError(excp) from excp

    @classmethod
    def _is_backend_working(cls, keyring_backend):
        """Check that a backend is working properly.

        It can happen so that a backend is installed but it might be
        misconfigured. But adding this test, we can asses if the backend
        is working correctly or not. If not then another backend should be tried instead.

        keyring.errors.InitError will be thrown if the backend system can not be initialized,
        indicating that possibly it might be misconfigured.
        """
        try:
            keyring_backend.get_password(
                "ProtonVPN",
                "TestingThatBackendIsWorking"
            )
            return True
        except (
            keyring.errors.InitError, keyring.errors.KeyringLocked,
            keyring.errors.NoKeyringError
        ):
            logger.exception("Keyring %s error", keyring_backend)
            return False



"""Module for Linux Secret Service Keyring backend.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import json
import os

import logging
import keyring
from proton.keyring.exceptions import KeyringLocked, KeyringError
from proton.keyring_linux.core import KeyringBackendLinux

logger = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class KeyringBackendLinuxSecretService(KeyringBackendLinux):
    """Implements the Secret Service keyring backend."""
    @classmethod
    def _get_priority(cls):
        return 5.

    @classmethod
    def _validate(cls):
        # this keyring doesn't work with snap containment
        if os.environ.get("SNAP") is not None:
            return False
        try:
            # pylint: disable=import-outside-toplevel
            from keyring.backends import SecretService
            return cls._is_backend_working(SecretService.Keyring())
        except ModuleNotFoundError:
            logger.debug("Gnome-Keyring module not found")
            return False

    def __init__(self):
        # pylint: disable=import-outside-toplevel
        from keyring.backends import SecretService
        self._backend = SecretService.Keyring()
        super().__init__(self._backend)

    def _get_item(self, key):
        try:
            stored_data = self._backend.get_password(
                self.KEYRING_SERVICE,
                key
            )
        except keyring.errors.KeyringLocked as excp:
            logging.info("Keyring locked while getting")
            raise KeyringLocked("Keyring is locked") from excp
        except keyring.errors.KeyringError as excp:
            logging.exception("Keyring error while getting")
            raise KeyringError(excp) from excp

        # Since we're borrowing the dict interface,
        # be consistent and throw a KeyError if it doesn't exist
        if stored_data is None:
            raise KeyError(key)

        try:
            return json.loads(stored_data)
        except json.JSONDecodeError:
            # gnome-keyring has a bug when processing new lines.
            # Anytime we store the data to the keyring, we store all newlines in
            # escaped form, aka \\n. For some reason, when gnome-keyring has no password
            # set, upon reading from the keyring it seems auto-escape the string and the
            # output becomes \n thus, treating this as a literal new line character,
            # causing a crash when we try to load the string via json, because again, \n are
            # treated literally. On the contrary, when the keyring has a password set,
            # and when we retrieve the data from the keyring, the new lines are properly
            # escaped and thus it does not crash once we try to load the string to json format.
            # From out testing, both via secret-storage and libsecret packages
            # (and even secret-tool) we were able to recreate this issue, thus indicating that
            # the issue is with gnome-keyring and not related to packages that
            # communicate with the gnome-keyring.
            # See more at:
            # https://discourse.gnome.org/t/possible-bug-or-feature-storing-getting-data-keyring-protected-vs-unprotected-keyring/20312

            # This fixed was found by Wizzerinus on GitHub and a PR was create by Anonymous941
            # https://github.com/ProtonVPN/python-proton-keyring-linux/pull/1

            stored_data = stored_data.replace("\n", "\\n")

        try:
            return json.loads(stored_data)
        except json.JSONDecodeError as excp:
            # Something is wrong here, delete data (it's invalid anyway)
            logging.exception("Keyring credential is not valid JSON, deleting")
            self._del_item(key)
            raise KeyError(key) from excp
